from fuzzywuzzy import fuzz, process


def default_process(query_value, block_values):
    return process.extractOne(query_value, block_values)


def partial_ratio_process(query_value, block_values):
    return process.extractOne(
        query_value, block_values, scorer=fuzz.partial_ratio
    )


def token_sort_ratio_process(query_value, block_values):
    return process.extractOne(
        query_value, block_values, scorer=fuzz.token_sort_ratio
    )


def token_set_ratio_process(query_value, block_values):
    return process.extractOne(
        query_value, block_values, scorer=fuzz.token_set_ratio
    )



# Generated by Django 2.0 on 2017-12-18 01:30

import crosswalk.models.user
import crosswalk.validators
from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ApiUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.SlugField(default=crosswalk.models.user.ApiUser.generate_token, editable=False, max_length=20, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Domain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField(blank=True, editable=False, max_length=250, unique=True)),
                ('name', models.CharField(max_length=250, unique=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='children', to='crosswalk.Domain')),
            ],
        ),
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(validators=[crosswalk.validators.validate_shallow_dict, crosswalk.validators.validate_no_reserved_keys])),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('alias_for', models.ForeignKey(blank=True, help_text='Entity in the same domain whose UUID supersedes this one.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='aliases', to='crosswalk.Entity')),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('domain', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='entities', to='crosswalk.Domain')),
                ('superseded_by', models.ForeignKey(blank=True, help_text='Entity in another domain whose UUID supersedes this one.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='supersedes', to='crosswalk.Entity')),
            ],
            options={
                'verbose_name_plural': 'entities',
            },
        ),
        migrations.AlterUniqueTogether(
            name='entity',
            unique_together={('domain', 'attributes')},
        ),
    ]



import random
import string

from django.contrib.auth.models import User
from django.db import models


class ApiUser(models.Model):
    def generate_token():
        return "".join(
            random.SystemRandom().choice(string.ascii_letters + string.digits)
            for _ in range(20)
        )

    user = models.OneToOneField(User, on_delete=models.CASCADE)
    token = models.SlugField(
        max_length=20, default=generate_token, editable=False, unique=True
    )

    def __str__(self):
        return self.user.username



from crosswalk.exceptions import NestedAttributesError, ReservedKeyError


def validate_shallow_dict(value):
    """Only allow shallow dictionaries."""
    if isinstance(value, dict):
        for key in value:
            if isinstance(value[key], dict):
                raise NestedAttributesError(
                    "Nested attributes are not allowed."
                )


def validate_no_reserved_keys(value):
    reserved_keys = [
        "alias_for",
        "aliased",
        "attributes",
        "created",
        "domain",
        "entity",
        "match_score",
        "superseded_by",
        "uuid",
    ]

    if isinstance(value, dict):
        for key in value:
            if key in reserved_keys:
                raise ReservedKeyError("Reserved key in attributes.")


def full_validation(value):
    validate_shallow_dict(value)
    validate_no_reserved_keys(value)



import uuid

from crosswalk.models import Domain
from crosswalk.validators import (
    validate_no_reserved_keys,
    validate_shallow_dict,
)
from django.contrib.auth.models import User
from django.contrib.postgres.fields import JSONField
from django.db import models


class Entity(models.Model):
    uuid = models.UUIDField(
        default=uuid.uuid4, editable=False, primary_key=True
    )

    domain = models.ForeignKey(
        Domain, related_name="entities", on_delete=models.PROTECT
    )

    attributes = JSONField(
        validators=[validate_shallow_dict, validate_no_reserved_keys]
    )

    alias_for = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        related_name="aliases",
        on_delete=models.CASCADE,
        help_text="Entity in the same domain whose UUID supersedes this one.",
    )

    superseded_by = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        related_name="supersedes",
        on_delete=models.SET_NULL,
        help_text="Entity in another domain whose UUID supersedes this one.",
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(
        User, related_name="+", null=True, on_delete=models.SET_NULL
    )

    @property
    def is_alias(self):
        return bool(self.alias_for)

    @property
    def is_superseded(self):
        return bool(self.superseded_by)

    def __str__(self):
        return self.attributes.get("name", self.uuid.hex)

    class Meta:
        verbose_name_plural = "entities"
        unique_together = ("domain", "attributes")



from rest_framework import status
from rest_framework.response import Response

from crosswalk.authentication import AuthenticatedView


class ClientCheck(AuthenticatedView):
    def get(self, request):
        return Response(
            "Client configured correctly", status=status.HTTP_200_OK
        )



from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers

from .models import Domain, Entity


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ("username", "first_name", "last_name", "email")


class DomainSerializer(serializers.ModelSerializer):
    name = serializers.CharField(validators=[])
    parent = serializers.SlugRelatedField(
        slug_field="slug",
        queryset=Domain.objects.all(),
        required=False,
        allow_null=True,
    )

    def create(self, validated_data):
        try:
            parent = Domain.objects.get(slug=validated_data.get("parent"))
        except ObjectDoesNotExist:
            parent = None
        domain, created = Domain.objects.get_or_create(
            name=validated_data.get("name"), defaults={"parent": parent}
        )
        return domain

    class Meta:
        model = Domain
        lookup_field = "slug"
        fields = ("slug", "name", "parent")


class EntityListSerializer(serializers.ListSerializer):
    def create(self, validated_data):
        entities = [Entity(**data) for data in validated_data]
        return Entity.objects.bulk_create(entities)


class EntitySerializer(serializers.ModelSerializer):
    uuid = serializers.UUIDField(format="hex_verbose", required=False)
    domain = serializers.SlugRelatedField(
        slug_field="slug", queryset=Domain.objects.all()
    )
    alias_for = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        required=False,
        allow_null=True,
        pk_field=serializers.UUIDField(format="hex_verbose"),
    )
    superseded_by = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        required=False,
        allow_null=True,
        pk_field=serializers.UUIDField(format="hex_verbose"),
    )

    class Meta:
        model = Entity
        list_serializer_class = EntityListSerializer
        fields = ("uuid", "attributes", "domain", "superseded_by", "alias_for")



from crosswalk.authentication import AuthenticatedView
from crosswalk.exceptions import NestedAttributesError, ReservedKeyError
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.validators import (
    validate_no_reserved_keys,
    validate_shallow_dict,
)
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response


class BulkCreate(AuthenticatedView):
    def post(self, request, domain):
        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response("Domain not found.", status=status.HTTP_200_OK)

        entities = request.data.copy()

        entity_objects = []

        for entity in entities:
            uuid = entity.pop("uuid", None)

            # Validate entity attributes before creating in bulk
            try:
                validate_shallow_dict(entity)
            except NestedAttributesError:
                return Response(
                    "Cannot create entity with nested attributes.",
                    status=status.HTTP_400_BAD_REQUEST,
                )
            try:
                validate_no_reserved_keys(entity)
            except ReservedKeyError:
                return Response(
                    "Reserved key found in entity attributes.",
                    status=status.HTTP_400_BAD_REQUEST,
                )

            entity_objects.append(
                Entity(
                    uuid=uuid,
                    domain=domain,
                    attributes=entity,
                    created_by=request.user,
                )
            )

        created_entities = Entity.objects.bulk_create(entity_objects)

        return Response(
            {
                "entities": [
                    {"entity": EntitySerializer(entity).data, "created": True}
                    for entity in created_entities
                ]
            },
            status=status.HTTP_200_OK,
        )



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response

from crosswalk.authentication import AuthenticatedView
from crosswalk.models import Domain, Entity


class DeleteMatch(AuthenticatedView):
    def post(self, request, domain):
        """
        POST searches for an entity based on criteria. If only one entity is
        returned from query, it is deleted. If more than one, return 403.
        """
        block_attrs = request.data.copy()

        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.", status=status.HTTP_404_NOT_FOUND
            )

        entities = Entity.objects.filter(domain=domain)
        entities = entities.filter(attributes__contains=block_attrs)

        if entities.count() == 0:
            return Response(
                "Entity not found.", status=status.HTTP_404_NOT_FOUND
            )

        elif entities.count() > 1:
            return Response(
                "More than one entity found.", status=status.HTTP_403_FORBIDDEN
            )

        entities.delete()

        return Response(status=status.HTTP_204_NO_CONTENT)



from crosswalk.authentication import AuthenticatedView
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.utils import import_class
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response


class BestMatch(AuthenticatedView):
    def post(self, request, domain):
        """
        Get the best matched entity for a given query.

        If the entity is an alias of another entity, the aliased entity is
        returned.
        """
        data = request.data.copy()
        query_field = data.get("query_field")
        query_value = data.get("query_value")
        return_canonical = data.get("return_canonical", True)
        block_attrs = data.get("block_attrs", {})
        scorer_class = data.get("scorer", "fuzzywuzzy.default_process")

        try:
            scorer = import_class("crosswalk.scorers.{}".format(scorer_class))
        except ImportError:
            return Response(
                "Invalid scorer.", status=status.HTTP_400_BAD_REQUEST
            )

        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.", status=status.HTTP_404_NOT_FOUND
            )

        entities = Entity.objects.filter(domain=domain)
        entities = entities.filter(attributes__contains=block_attrs)

        entity_values = [e.attributes[query_field] for e in entities]

        match, score = scorer(query_value, entity_values)

        entity = entities.filter(
            **{"attributes__{}".format(query_field): match}
        ).first()

        aliased = False

        if return_canonical:
            while entity.alias_for:
                aliased = True
                entity = entity.alias_for

        return Response(
            {
                "entity": EntitySerializer(entity).data,
                "match_score": score,
                "aliased": aliased,
            },
            status=status.HTTP_200_OK,
        )



import os

from setuptools import find_packages, setup

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-crosswalk",
    version="0.0.9",
    packages=find_packages(exclude=("example",)),
    include_package_data=True,
    license="MIT",
    url="https://github.com/The-Politico/django-crosswalk",
    description="Build and query databases of arbitrary entities.",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    install_requires=[
        "django",
        "djangorestframework",
        "django-filter",
        "django-uuslug",
        "python-Levenshtein",
        "fuzzywuzzy",
    ],
)
